from __future__ import division
import os,sys,time,new
from cuisine.pour import deck
from cuisine import dvfile,dvframe
from louie import dispatcher
from socket import getfqdn
from cuisine.descnodes import *
from cuisine.desctypes import *
from cuisine.utils import Condition
import logging
log=logging.getLogger("Recorder")
tlog=logging.getLogger("timing")

"""
todo:

if recorder sees the timecode jump, it should request that the deck
back up and get the frames again

recorder is _far too big_! i need to write desc files based on frame
data, write new movie files with or without desc files, write movie
files in random access ways, etc. Recorder needs to be split into more
pieces to allow that

"""

class Descfileaccess:
    """mixin for recorder; these are the methods that access the desc files"""
    def startdescfile(self):
        """starts a new desc file for the current filename and saves it"""
        self.descfile=DescFileManager().newfileforclip(self.filename)
        self.initclip()
        self.descfile.save()
        
    def initclip(self):
        """starts up a new clip with as much data as we can come up with"""
        clip=self.descfile.clip()
        ess=clip.essencefile()
        env=ess.creator().environment()
        env.user(os.environ.get('USER','unknown'))
        env.host(getfqdn())
        env.time(Timestamp())
        
        ess.creator().commandline(" ".join(sys.argv))


    def firstframedata(self,frame):
        """update the clip with information that can be determined from the first frame.
        """
        clip=self.descfile.clip()
        m=clip.essencefile().media()
        m.video().framesize().widthheight(*dvframe.getwidthheight(frame))

        # i don't actually know the creator of the dv; the deck i use
        # to play it back is not important at all (although we could
        # use libraw1394 to get info about that device)
        self.descfile.save()
    def laterframedata(self,frame,framecount):

        thistc=dvframe.gettimecode(frame)
            
        # update the file duration field
        ess=self.descfile.clip().essencefile()
        video=ess.media().video()
        video.duration(Duration(framecount,"frames"))

        thisframenum=Duration(self.recordedframecount-1,"frames") # ? not sure why
    
        # update the timecode channel
        if thistc:
            tc=ess.channels().byparamname("timecode")

            tc.setpoint(thisframenum,Duration(thistc))
        # and recorddate channel
        #-- not ready yet since timestamps don't have math
        #rd=self.deck.getrecorddate()
        #if rd:
        #    c=ess.channels().channel("recorddate")
        #    c.setpoint(framenum,rd)
        
        # and atn channel
        if self.deck:
            atn=self.deck.getATN()
            if atn:
                if 0: # the approximate ATN channel, disabled since it doesnt work
                    ac=ess.channels().findparam("absoluteTrackNumber")
                    if ac is None:
                        # setup new atn Channel
                        ac=ess.channels().byparamname("absoluteTrackNumber")

                        # we dont want to record the irregularities in the
                        # ATN- it's already not frame-accurate, so storing
                        # the noise is a waste.

                        # in my test, atn varied from linear by -4..10 atn
                        # units. there are just about 10 units per frame i only
                        # mean to set this param once.

                        ac.maximumerror(55) 

                    ac.setpoint(thisframenum,atn)
                # a testing channel with -all- the atn data on it
                ac2=ess.channels().byparamname("absoluteTrackNumber-exact")
                ac2.setpoint(thisframenum,atn)

    def savedesc(self):
        self.descfile.save()
        
class Recorder(Descfileaccess):
    
    """recorder gets frames from a Deck and puts them in its own file
    on disk.  recorder will not write consecutive frames with the same
    timecode (unless the timecode is None); it ignores
    repeats. recorder also writes a descfile as it records the frames.

    recorder watches out for existing files.

    if nofilesout is set, this recorder will skip the creation of the
    essence file (but it will still attempt to create the matching
    desc file as usual). this option is used in the 'baste' utility to
    create a desc file for a foreign essence file. Also, setting
    nofilesout supresses the frames-paused detection, since presumably
    it's already known which frames are in the datafile; it will just
    corrupt things for the recorder to try to skip some.

    signals:
      'recorder status', msg, internalmsg
        this gets sent when the recorder has something to say about a
        recording getting started; not working; etc. it gets logged at
        level INFO and also gets sent as a signal. see logandstatus()
        for the arguments

    todo:
       announce when frames start and stop arriving
    
    """
    rolling=0 # are we currently putting new frames to the disk?
    deck=None
    filename=None # the filename we're recording to now, or will record to next
    recordfile=None # the file object we're writing to
    statuscb=lambda *args: None # callback to receive status messages (probably the gui)
    lasttimecode=None

    def __init__(self,nofilesout=0):
        self.timecodepaused=Condition(truemsg="frames paused, last timecode is %s",
                                      falsemsg="unpaused, timecode resumes from %s",log=log)
        self.framelimit=None
        self.nofilesout=nofilesout
        self.deck=None
    def setdeck(self,d):
        self.deck=d # recorder needs to be able to ask deck for frame callbacks
    def setformat(self,format):
        """tell the recorder which fileformat in ('dv1','dv2','raw')
        to use on new files. appending to existing files should be
        done in their existing format"""
        self.format=format
    def setrecording(self,rolling):
        """use this to start/stop recording.
        returns 0 if your request failed, 1 for success"""
        if rolling>self.rolling:
           return self.startrecording()
        if rolling<self.rolling:
           return self.stoprecording()
        # no change to record state:
        return 1
    def setfile(self,filename):
        """if we're not recording: choose the filename for the next
             recording
           if we're recording: change the filename of the current
             record (by moving the file on disk)
        """
        filename=str(filename)
        if not self.isrolling():
            log.debug("recorder to use filename: %s%s" % (filename,
                                              ('',' (writing desc only)')[self.nofilesout]))
            self.filename=filename
            return 1
        else:
            # change name while rolling - we move to the new file and
            # resume there
            if self.nofilesout:
                log.warn("it's odd that you'd change the output filename when the recorder"+
                         "isn't set to output an essence file - ignoring name change")
                return 0
            log.warn("tried to change filename while recording - not supported yet")
            return 0
    def setframesperfile(self,limit=None):
        
        """if limit is set, recordings will self-stop after that
        number of frames (per recording- not a global count). defaults
        to None, which means no limit."""

        self.framelimit=limit
        
    def newframe(self,frame):
        """callback. the deck calls this on each new frame"""
        if not self.isrolling():
            return

        lasttc=self.lasttimecode

        thistc=dvframe.gettimecode(frame)

        # this feature makes the recorder stop when the timecode
        # is repeating. there's no effect if the timecode is missing
        # (such as in a live capture on my camera), and there's no
        # effect if nofilesout flag is set.
        if thistc and not self.nofilesout and (lasttc==thistc): 
            # repeated frame
            self.timecodepaused.set(1,thistc)
            return
        
        t=time.time()
        # new frame
        self.timecodepaused.set(0,thistc)
        self.lasttimecode=thistc

        if not self.nofilesout:
            tlog.debug("  write frame %.02f" % ((time.time()-t)*1000))
            self.recordfile.writeframe(frame)
        self.recordedframecount+=1

        tlog.debug("  xml %.02f" % ((time.time()-t)*1000))
        self.laterframedata(frame,self.recordedframecount)

        # initialize the static stuff, if this is a new file
        if self.recordedframecount==1:
            self.firstframedata(frame)

        # periodic saves of the xml
        if (self.deck and self.recordedframecount%60==0) or (not self.deck and self.recordedframecount%500==0):
            tlog.debug("  save xml %.02f" % ((time.time()-t)*1000))
            self.savedesc()

        if self.framelimit and self.recordedframecount>=self.framelimit:
            # auto-stop
            log.info("recorder stopped at %i frames" % self.recordedframecount)
            self.stoprecording()

        tlog.debug("  done %.02f" % ((time.time()-t)*1000))

    def recordedframes(self):
        "returns how many frames have been written into this file"
        return self.recordedframecount
    def isrolling(self):
        return self.rolling

    ##############################################################
    def logandstatus(self,msg,internalmsg):
        # the internalmsg shall be forwaredd to other program modules,
        # which will act on it. so don't ever change the wording on
        # internalmsg, although you can change msg if you want.
        log.info(msg)
        dispatcher.send("recorder status",msg=msg,internalmsg=internalmsg) # old-style, rarely used
        dispatcher.send(signal=internalmsg,sender="recorder status",msg=msg) # new-style, use this

    def startrecording(self):
        """opens file and starts recording. setrecording ensures that
        we were not already rolling"""

        if not self.nofilesout:
            if os.path.exists(self.filename):
                log.error("file '%s' exists - won't overwrite",self.filename)
                return 0
            try:
                self.recordfile=dvfile.openForWrite(self.filename,self.format)
            except Exception,e:
                self.logandstatus("Can't open file for writing: %s"%e,"openfail")
                return 0

            self.logandstatus("Recording started to file %s" % self.filename,"startrecord")
        else:
            self.logandstatus("Desc file creation started for file %s" % self.filename,"startrecord")
            
        # tell the deck to call us with new frames
        if self.deck:
            self.deck.addnewframecb(self.newframe)
        # else:
        #   the owner of this Recorder will need call newframe() with the frames

        # reset our own counters and flags
        self.recordedframecount=0
        self.lasttimecode=None
        self.timecodepaused.clear()

        # start a new desc file for this new dvfile
        self.startdescfile()
        
        self.rolling=1
        return 1

    def stoprecording(self):
        """stop recording and close the file. setrecording() ensures
        that we were indeed rolling and had a file open"""

        # turn off the callbacks from the deck
        if self.deck:
            self.deck.delnewframecb(self.newframe)
        if not self.nofilesout:
            self.recordfile.closewrite()
        self.rolling=0
        
        self.savedesc()
        self.logandstatus("Recording stopped","stoprecord")


