from __future__ import division, nested_scopes, generators
from cuisine.xmlnodebase import xmlnodeclass,xmldocfile,collectiveelement
import os,new,bisect
from louie import dispatcher
from cuisine.desctypes import *
from cuisine.utils import *
from cuisine.external.dictmixin import DictMixin
from cuisine.descnodes import Clip, DescFileManager, Channels, Channel

import logging
log=logging.getLogger("shownodes")

"""
this module is for import *

todo:
clip, track, show need a base type (<-- i have forgotten why, since writing that node)
"""


class Source(xmlnodeclass):
    
    """some bricks has a source, which involves a filename (timestamed
    reference!), a component from that file, and maybe more. the
    filename is in the text node of the source; the rest is in
    attributes.

    the filename is the desc filename. old bricks may list the media
    filename instead.

    which brick types have source files?

        clip - always
        effect - maybe, if the effect has been cached to a file
        show - possible, if the show has been cached to a file
        clone - maybe

        source is stored as a filename, although for convenience you can
        pass a Clip in and we'll extract the filename for you.


    """
    def filename(self,newval=None):
        if newval is None:
            # get
            # get the text within this <source>fname</source> node
            return self._getorsettextfornode(node=self)

        # set
        if isinstance(newval,Clip):
            # auto-extract filename from clip
            newval=newval.essencefile().filename()            
        ret=self._getorsettextfornode(self,newval)

        if newval is not None:
            # in a 'set' action, we also stamp the element
            self.setAttribute("time",str(Timestamp()))

        return ret

    #######################################

    def descfile(self):
        "gets the Descfile for the clip."
        dfm=DescFileManager()
        f=self.filename()
        try:
            return dfm.openfilebydescfilename(f)
        except IOError:
            # for old-style bricks that use media filename:
            return dfm.openfileforclip(f)  

    def tbd_getframe(self,t,component):
        
        """returns the media for the given frame (or None if no media
        applies). media might be a Frame or an AudioFrame."""

        df=self.descfile()
        ef=df.clip().essencefile()
        ret=df.clip().essencefile().media().tbd_getframe(t.getas("frames"),component)
        return ret
   


class Brick(xmlnodeclass):
    #
    # marshaling code first
    #
    def __init__(self,bricktype="clip",**kw):

        """ when making a brick, you can include keywords named after
        method calls you want run on the new brick.  like,
          b=Brick(origin=x,intime=y)
        is the same as
          b=Brick(); b.origin(x); b.intime(y)
        """
        
        # this has to go before the super init, since that calls readxml, which
        # wants to operate on the brick type

        xmlnodeclass.__init__(self)
        self.bricktype(bricktype)
        self.assignuuid()

        # use any kw
        for k,v in kw.items():
            meth=getattr(self,k)
            meth(v)

    def bricktype(self,newval=None):
        if newval:
            # make sure we mix in the other type's stuff
            self.changetype(newval)
        return self._getorsetattr("type",newval)

    def readxml(self):
        """very tricky. we look at the type= attribute of the brick
        and strap on any relevant mixin classes (if we haven't already)."""

        for a in 'cache_origin','cache_intime','cache_outtime':
            if hasattr(self,a):
                delattr(self,a)

        self.changetype(self.bricktype())
        
    def changetype(self,typename):
        
        """typename is like 'edit' or 'effect' (or 'clip', which is
        ignored). causes a brick to change in-place. probably you want
        this only if you just made the brick object. it's not called
        by __init__ because i'm too scared about what it will
        break. calling with the basic clip type will have no bad
        effect. also, i hope this is idempotent (probably is)."""

        from cuisine.shownodes.edit import EditBrickMixin
        from cuisine.shownodes.effect import EffectBrickMixin

        mixintable={"edit":EditBrickMixin,"effect":EffectBrickMixin}
        if typename in mixintable:
            mixinclass=mixintable[typename]
            
            # this is a specialized brick, so we forcibly mix in the
            # given mixin class. i'm just replacing/adding methods;
            # not changing the type hierarchy, etc, so all sorts of
            # bad things could happen
            # Further reading:
            #   http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/81732
            #   http://www.python.org/doc/current/lib/module-new.html
            for n in mixinclass.__dict__:
                if not n.startswith('__'): # skip __python__ methods
                    othermeth=mixinclass.__dict__[n]
                    bound=new.instancemethod(othermeth,self,self.__class__)
                    self.__dict__[n]=bound

            log.debug("brick is type '%s', added %r mixin" % (typename,mixinclass.__name__))

    def origin(self,newval=None):
        """the origin of the brick, relative to the start of the show"""
        if newval is None:
            ### value cache
            if not hasattr(self,'cache_origin'):
                self.cache_origin=self._getorsettypednode("origin",Duration)
            return self.cache_origin
            ###
        else:
            dispatcher.send("bricks moved",brick=self)
            if hasattr(self,'cache_origin'):
                del self.cache_origin
            return self._getorsettypednode("origin",Duration,newval)

    def intime(self,newval=None):
        """the time that the brick starts, relative to the brick's origin.
        intime and outtime are like crop or trim points. moving them doesnt
        affect the body of the clip (since the origin is unchanged). if
        intime isn't set, it will get set to 0 when accessed.
        """
        if newval is None and hasattr(self,'cache_intime'):
            return self.cache_intime
        
        if newval is not None:
            # set
            dispatcher.send("bricks moved",brick=self)
            if newval.getas("frames")<0:
                newval=Duration(0,"frames")
                
        ret=self._getorsettypednode("intime",Duration,newval)
        if ret is None:
            ret=self._getorsettypednode("intime",Duration,Duration(0,"frames"))
            
        self.cache_intime=ret
        return ret
    
    def outtime(self,newval=None,checkcliplength=1):
        """the time the brick ends, relative to the brick's origin.
        unset outtimes for clip bricks will get set to the clip's length.
        unset outtimes for other nodes will not get set (they'll return None
        if you access them).

        if this clip has a source, we will clamp the outtime to the length of the source clip.

        if checkcliplength is not set, the clip will not be checked for length.
        """

        if newval is None and hasattr(self,'cache_outtime'):
            return self.cache_outtime

        if newval is not None:
            if checkcliplength and self.source():
                df=self.source().descfile()
                cliplength=df.clip().essencefile().media().video().duration()
                log.debug("compare %r %r" % (newval,cliplength))
                if newval > cliplength:
                    newval=cliplength
        
        ret=self._getorsettypednode("outtime",Duration,newval)

        if ret is None:
            if self.source():#self.bricktype=='clip':
                # go find the clip's length for use as a default outtime
                log.debug("setting the empty outtime to the brick's source's duration")
                df=DescFileManager().openfileforclip(self.source().filename())
                cliplength=df.clip().essencefile().media().video().duration()
                ret=self._getorsettypednode("outtime",Duration,cliplength)
            else:
                log.warn("brick %s has no outtime, and no source for us to use to make one up" % self)

        if newval is not None:
            dispatcher.send("bricks moved",brick=self)

        self.cache_outtime=ret
        return ret
    def track(self,newval=None):
        """the track attribute is a Track object that this brick sits on
        (someday, bricks may be able to hop tracks. but not for now). in the xml,
        tracks are stored by name, but this function operates on Track objects."""

        if newval is None and hasattr(self,'cache_track'):
            return self.cache_track
        
        # we convert between track name and object right in the accessor here
        if newval is None:
            # get
            trackname=self._getorsetstringnode("track")
            try:
                ret=self.parentbytype(Show).tracks().byname(trackname)
            except AttributeError:
                log.error("couldn't lookup track named '%s' on brick '%s'" % (trackname,self))
                ret=None
        else:
            # set
            self._getorsetstringnode("track",newval.name())
            dispatcher.send("bricks moved",brick=self)
            ret=newval
        
        self.cache_track=ret
        return ret

    def source(self,newval=None):
        """see the Source docstring for more. you can test this return value to
        find out if the brick has a source (you won't create one)"""
        
        return self._getorset1(Source,newval)

    def channels(self,newval=None):
        """the brick might have a channels node. this accessor will not create
        a channels block, so you can use it to query whether there are channels or not.
        (like with source()).
        """
        return self._getorset1(Channels,newval)

    #
    # now the Brick methods
    #
    def __str__(self):
        s=self.source()
        if s:
            # nodes with sources can present that (with component) "/path/to/clip (video)"
            return "%s" % (s.filename())
        else:
            # other nodes appear as their bricktype "[edit]"
            return "[%s]" % (self.bricktype())
        
    
    def duration(self,newval=None):
        """the duration of the brick, i.e. the brick end time minus
        the intime.  the duration is really calculated from outtime()-intime(),
        but you can still assign to duration. adjusting the duration
        changes the outtime (an arbitrary decision of mine).  """
        if newval is None:
            return self.outtime()-self.intime()
        else:
            self.outtime(self.intime()+newval)
            return newval

    def isactiveattime(self,t):
        """returns whether or not this brick's span includes the given time,
        which is measured from the start of the show. note that this doesn't
        mean you have a frame yet; this brick could reference another track which
        has no brick at the time."""
        t1=self.intime()
        t2=self.outtime()
        o=self.origin()
        if t1 is None or t2 is None or o is None:
            raise ValueError("brick is missing intime, outtime, or origin, can't check range")
        ret = t1<=(t-o)<=t2
        return ret

    def clip(self):
        "shortcut to the clip brick's clip, if it has one. returns null otherwise"
        s=self.source()
        if s is None:
            return None
        
        return s.descfile().clip()
            
    def tbd_getframe(self,t,component):
        """returns the frame at time t, measured from the brick's time
        origin. perhaps the frame is fetched from a clip file, perhaps
        it's created by an effect program. if there is no frame data
        for that component for some reason-- including that the brick
        doesn't cover the given time, or that that the effect doesn't
        output the given component-- this method shall return None,
        which is probably interpreted as silence or black

        i dont like this api, since it's not friendly toward 
        alternate timing systems. that's why there's tbd_ on it.

        component is 'video' or 'audio'.

        this version is for clips and other things with sources. effect
        and edit bricks replace this method.
        """
        if t<self.intime() or t>self.outtime():
            # this is ok; bricks return None frames outside of their range
            return None

        if self.source() is None:
            return None

        return self.source().tbd_getframe(t,component)
    def getothertrackbrick(self,track,t,returntime=0):
        """
        this method and the next two are for edit/effect bricks that
        need to refer to other bricks. clip bricks won't use this
        message. the t argument is always time within this brick.

        method for getting a frame from another track at the same time
        as the t in this brick. you pass t measured from this brick's
        origin, and this method will go to the other track and run
        tbd_getframe on the brick that's there (at the correct time
        for that brick).

        returns None if there was no frame for any reason (like the
        time was out of range for this brick or the other brick)

        if the returntime arg is true, the return value will be a
        tuple of the other brick and the time in that brick's time
        system.

        
        times aren't autoconverted yet, so here's how the time is handled:

           [e0      et   ] <-- this edit brick
               [c0  ct        ] <-- the input brick

        e0 is this edit brick's origin; et is the given time t
        c0 is the input brick's origin; ct is a time in that brick

        ct = et+e0-c0 = et-(c0-e0)

        ## due to additional changes, all these methods should be revamped
        ## to properly make use of each other again (and not repeat code)
        """
        if not self.intime()<=t<=self.outtime():
            return None

        showtime=t+self.origin()
        otherbrick=self.parentbytype(Bricks).gettrackbrick(track,showtime)
        if not returntime:
            return otherbrick
        else:
            return otherbrick,otherbrick and showtime-otherbrick.origin()
    def getotherbrickframe(self,brick,t,component):
        if not self.intime()<=t<=self.outtime():
            return None
        
        if brick is None:
            return None
        showtime=t+self.origin()
        otherbricktime=showtime-brick.origin()
        return brick.tbd_getframe(otherbricktime,component)

    def getothertrackframe(self,track,t,component):
        if not self.intime()<=t<=self.outtime():
            return None

        if getattr(self,'_norecurse',0):
            log.warn("edit brick on %s refers to itself",self.track())
            return None
        self._norecurse=1
        b=self.getothertrackbrick(track,t)
        ret=self.getotherbrickframe(b,t,component)
        self._norecurse=0
        return ret

    def sliceadjusttimes(self,t,newbrick):
        """part of the work of a slice is to adjust the in/out times of the copy.
        t is the requested time of the slice."""
        newbrick.intime(t)
        newbrick.outtime(self.outtime())
        self.outtime(t-Duration(1,'frames'))

    def slice(self,t):
        """
        this is harder than it sounds.

        this method slices the brick into two at time t (measured from
        this brick's origin). the show is guaranteed not to look any
        different after the slice. however, it is not guaranteed that
        you can delete the right half and extend the left half's
        outpoint and get to the same state you were in before. that's
        true for clip bricks and maybe effect bricks, but not for edit
        bricks (since the cuts move to the new brick).

        the present brick is modified to become the left half of the
        result. a new brick is created and inserted into the show. the
        new brick is also returned from this method.
        
        """
        log.info("slicing brick %s" % self)
        show=self.parentbytype(Show)

        b2=self.copy()
        show.bricks().add(b2)

        self.sliceadjusttimes(t,b2)

        # channels should be handled too!
        
        return b2

    def addaudiocurve(self):
        """this is a temporary hack that gives the brick a 1.0 audio
        curve from its inpoint to its outpoint, with two handles
        within the curve"""
        if not self.channels():
            log.info("making new channels block on %s" % self)
            self.channels(Channels())

        chan=self.channels().byparamname('audioLevel')
        if chan.getsegments():
            log.warn("audioLevel param already existed - won't touch it")
            return

        t1=self.intime()
        t4=self.outtime()
        tt=(t4-t1)
        t2=t1+Duration(int(tt._val*.15),tt._unit) ### Unum needs fixing
        t3=t1+Duration(int(tt._val*.85),tt._unit)

        for xs in (t1,t2),(t2,t3),(t3,t4):
            seg=chan.addsegment()
            seg.x1,seg.x2=xs
            seg.y1=seg.y2=1        

class Bricks(collectiveelement):
        
    def childtype(self):
        return Brick

    def readxml(self):
        self.brickmap={}
        dispatcher.connect(self.bricksmoved,"bricks moved") # this shouldn't actually be repeated on each read
        dispatcher.send("bricks moved")
        self.buildbrickmap()
    def brickmap_rowtimings(self):
        return range(0,51200,100) # these were never used, so we dont have to construct Unums!
        #        return [Duration(x,"frames") for x in range(0,51200,100)] # about 15m of times, spread by about 3s
    def brickmap_indexfortime(self,t):
        """given a time, we return the greatest column number in the
        brickmap has a time left of t (or else 0)"""
        frames=t.getas("frames")
        col=int(frames//100)
        col=min(511,max(0,col))
        return col

    def brickmap_singlebrick(self,b,action='insert'): # or 'remove'
        """internal func for buildbrickmap, brick_premove, etc.
        if action is 'remove', the brick must be in the map already or there's
        an exception"""
        t=b.track()
        if t not in self.brickmap:
            # initialize a row of times
            self.brickmap[t]=[(x,[]) for x in self.brickmap_rowtimings()]
        row=self.brickmap[t]

        startcol=self.brickmap_indexfortime(b.intime()+b.origin())
        endcol=self.brickmap_indexfortime(b.outtime()+b.origin())

        for col in range(startcol,endcol+1):
            if action=='insert':
                row[col][1].append(b)
            else: # remove
                try:
                    row[col][1].remove(b)
                except ValueError:
                    raise ValueError("brick wasn't in brickmap as expected")

        
    def buildbrickmap(self):
        """
        this scans all bricks and stores them in a map described below.
        the point is to speed Show.gettrackbrick() as much as possible. you need
        to send the 'bricks moved' signal if any brick changes coordinates or tracks.

        brickmap=
            track1 : { (time1, [bricks that overlap the span from time1..time2-1]),
                       (time2, [bricks that overlap the span from time2..time3-1]),
                      ...
                       (time4, [bricks that cross the space after time4])
                      },
            track2 : { ...

        now it's quick to get the brick for a track-time, and pretty quick to get
        all bricks for a time (assuming small number of tracks).

        how are the time1, time2, etc decided? it doesn't matter. they should span the
        show, and be as dense as desired. they should be evenly spaced, to make the
        access even easier.

        yes, we could leave out the first element of the tuples, but a quick speed
        test suggests they don't affect performance, and future versions may want
        uneven time arrangements.
            
        """

        self.brickmap={}
        log.debug("build brickmap")
        show=self.parentbytype(Show)

        for b in self.getall():
            self.brickmap_singlebrick(b)
            

    def bricksmoved(self):
        """
        this means the brick map is invalid. if a brick is provided,
        then only that brick has moved, and the move took place before
        this method was called. someday, the algorithm may do
        incremental updating based on the brick argument.
        """
        log.debug("clear brickmap")
        self.brickmap.clear()


    # untested:
    def brickmap_premove(self,brick):
        """you can skip sending 'bricks moved' if you call this method
        on the brick that's about to change, and then call the next
        method right after the change."""

        self.brickmap_singlebrick(b,action='remove')
    def brickmap_postmove(self,brick):
        """see previous method"""
        self.brickmap_singlebrick(b,action='insert')
    
    def getbricksattime(self,t):
        """returns all the bricks at the current time coordinate
        as a dict of track:brick"""

        log.debug("getbricksattime %r",t)
        
        
        res={} # track name : brick object
        for b in self.getall():
            try:
                if b.isactiveattime(t):
                    res[b.track()]=b
            except ValueError,e:
                log.warn("brick %r skipped: %s" % (b,e))

        return res
    
    def gettrackbrick(self,track,t):
        """returns the brick that crosses the given Track at time t
        (measured from start of show). returns None if there's no
        brick there"""

        if not self.brickmap:
            log.debug("rebuild brickmap")
            self.buildbrickmap()

        if track not in self.brickmap:
            # there are no bricks on that track at all
            return None
        row=self.brickmap[track]
        i=self.brickmap_indexfortime(t)
        ret=None
        for b in row[i][1]: # these bricks cross the time segment

            if b.isactiveattime(t):
                log.debug("brickmap finds %s at t=%r tr=%r" % (b,t,track))
                return b
        return None

        
    def remove(self,brick):
        collectiveelement.remove(self,brick)
        self.bricksmoved()
        
class Track(xmlnodeclass):
    def name(self,newval=None):
        return self._getorsetattr("name",newval)
    def getframe(self,t,component):
        """returns the frame for time t on this track, or None if the track doesn't
        lead to any frame. component is 'video' or 'audio', depending on which sort
        of data you want. (you have to make two calls to fetch both components)"""
        # the track currently doesn't know what bricks are on it- we
        # ask the show for that
        return self.parentbytype(Show).gettrackframe(self,t,component)
    def __str__(self):
        return "<track %s>" % self.name()
    __repr__=__str__

class Tracks(collectiveelement):
    def childtype(self):
        return Track
    def byname(self,trackname):
        "returns the track object with the given name"
        matches=[x for x in self.getall() if x.name()==trackname]
        if not matches:
            raise ValueError("no track with name '%s'" % trackname)
        return matches[0]

class Show(xmlnodeclass):
    """show is a flat list of bricks in no ord"""
    #timeline layouts,
    #
    def __init__(self):
        xmlnodeclass.__init__(self)
        self.assignuuid()
    def readxml(self):
        self.cache=ClassyDict()
    def tracks(self,setnode=None):
        return self._getsetorcreate1(Tracks,setnode)
    def bricks(self,setnode=None):
        return self._getsetorcreate1(Bricks,setnode)

    def bgcolor(self,newval=None):
        """optional tint color attribute for the timeline display. use this
        to distinguish your different shows"""
        return self._getorsetattr("bgcolor",newval) or "#cccccc"

    ##############################################################
    def trackbyname(self,trackname):
        "simple shortcut"
        return self.tracks().byname(trackname)
  
    def gettrackframe(self,track,t,component):
        """returns the frame from the brick at time t (show time) on
        the given Track. returns None if there's nothing there"""
        b=self.bricks().gettrackbrick(track,t)
        if b is None:
            return None
        else:
            bricktime=t-b.origin()
            log.debug("fetching frame %r on brick %r" % (bricktime,b))
            return b.tbd_getframe(bricktime,component)


################################################################################
class ShowFile(xmldocfile):
    """a show file on disk, from which you can access the Show object."""
    def __init__(self,showfilename):

        self._openornew(showfilename,topleveltype=Show)

    def show(self):
        "get the Show that's the top node of the file"
        return self._gettoplevel()



        
try:
    import psyco
    psyco.bind(Bricks.buildbrickmap)
except ImportError:
    pass
