
from __future__ import nested_scopes,division
import gtk
import gtk.gdk as gdk
import gnomecanvas as canvas
keys=gtk.keysyms
from cuisine.shownodes import *
from cuisine.utils import *
from math import exp
from time import time
import sys
from skim.zooming import Pair
from cuisine.toss import pygtkutil
from louie import dispatcher
import logging
log=logging.getLogger("timeline")
flog=logging.getLogger("timeline-frequent") # events that happen, say, on every mouse motion

from brickview import Brickview
from gtkutils import *
from canvasobj import Canvasobj

class QuickEntry(gtk.Entry):
    """a slightly better entry widget that autoselects its contents
    when you focus into it."""
    def __init__(self,*a,**kw):
        gtk.Entry.__init__(self,*a,**kw)
        self.addhandlers()

    def addhandlers(self):
        """this makes it so the text is selected whenever you focus in
        to the entry widget. to make this work with mouse clicks, a
        mouse click signal that *causes* focus_in will be killed after
        a grab_focus() so that it doesn't alter the selection too"""
        
        def autoselect(wid,ev):
            wid.select_region(0,len(wid.get_text()))
            return 0
        self.connect('focus_in_event',autoselect)
        
        def focusbutdontselect(wid,ev):
            if not wid.is_focus():
                wid.grab_focus()
                return 1 # don't proceed to lose the selection
            return 0
        self.connect("button_press_event",focusbutdontselect)


class Cutpoint(Canvasobj):
    """a view for one of the cut points on an edit brick view, with its track
    entry box.

    you make this object on a particular cut point, although it will
    alter that cut point if the user drags the cut line.

    sensitive sets whether or not the cut line will be created. use 0
    for the first cut in the brick. it's not perfect yet; the first
    cut should probably move with the inpoint. (and we should stop
    allowing cuts to exist beyond the in/out points)

    """
    def __init__(self,master,brick,cutpoint,sensitive=1):
        make_attributes_from_args('master','brick','cutpoint')
        # note we have no link back to the editbrickview

        self.ids=ClassyDict()
        ids=self.ids

        ids.ent=self.master.makewidget(self.maketrackentry(),anchor="n")

        if sensitive:
            def onpress(drag,wid,ev):
                drag.startpoint=self.cutpoint
                drag.lastpos=self.cutpoint

            def onmotion(drag,wid,ev,delta):
                mas=self.master
                # determine desired new cut point from mouse motion
                self.altercut(drag.startpoint+mas.dtimefordxcoord(delta.x))
                # that may or may not have actually altered the cut. watch the
                # log for warnings.

                dispatcher.send("timeline changed")
                self.update()
            def onrelease(drag,wid,ev):
                self.master.update()


            ids.line=self.master.makeline()
            setproperties(ids.line,line_style=gtk.gdk.LINE_ON_OFF_DASH)

            ids.line.connect('event',self.mousedragfactory(button=1,
                                                           onpress=onpress,onmotion=onmotion,
                                                           onrelease=onrelease))

            ids.line.connect('event',self.highlightpropertyfactory("width_pixels",4))
        else:
            ids.line=None


    def altercut(self,newcutpoint):
        """this is the only method that will alter the cutpoint in the brick
        cuts table. returns 1 on successful adjustment of the brick cutpoint
        """
        
        # edit the brick's cut point list
        cuts=self.brick.cuts()
        if self.cutpoint not in cuts:
            log.warn("tried to alter cut point %s, which is not in the brick cut table" % cutpos)
            return 0
        if newcutpoint in cuts:
            log.warn("tried to alter a cut point over another one - not allowed")
            return 0
        cuts[newcutpoint]=cuts[self.cutpoint]
        del cuts[self.cutpoint]

        # change this object's cut point to our new key
        self.cutpoint=newcutpoint

        return 1
        
    def maketrackentry(self):
        """this makes an Entry widget that, when modified, will
        alter the track object in self.brick.cuts()[self.cutpoint] """

        # make a entry widget for the track number
        ent=QuickEntry()
        ent.set_has_frame(1)
        ent.set_width_chars(3)
        ent.show()

        ent.set_text( self.brick.cuts()[self.cutpoint].name() )

        def changed(wid):
            newtrackname=wid.get_text()
            # find the Track with the new name
            try:
                newtrack=self.brick.parentbytype(Show).trackbyname(newtrackname)
            except ValueError:
                # probably they're in the middle of choosing a new track,
                # in which case we want to leave the Entry alone (and inconsistent
                # with the model) to give them a change to form a correct name.
                # in keeping with the dicteditor widget, the Entry should
                # change color while it's inconsistent. of course, this shouldn't
                # be an entry in the first place
                log.info("no track named %s" % newtrackname)
                return 0 
            
            log.debug("you changed cuttrack to track %r" % newtrack)
            
            # set the edit brick to cut to the new track
            brickcuts=self.brick.cuts()
            brickcuts[self.cutpoint]=newtrack

            # cause an update
            dispatcher.send("timeline changed")
            
        ent.connect_after("changed",changed)

        return ent

    def withdraw(self):
        for o in self.ids.get('ent',None),self.ids.get('line',None):
            self.master.removecanvasobj(o)


    def update(self):
        "positions the line and label for this cutpoint"
        origin=self.brick.origin()
        mas=self.master
        
        x1=mas.xcoordfortime(origin+self.cutpoint) # the cut point

        # now we find the time for the end of this cut, which may be awkward if the
        # outtime comes before the last cut, but i leave it for now
        rightedge=self.brick.cuts().cutafter(self.cutpoint) or self.brick.outtime()
        x2=mas.xcoordfortime(origin+rightedge)
        
        y1,y2=mas.ycoordsfortrack(self.brick.track())
        if self.ids.line:
            setproperties(self.ids.line,points=[x1,y1,x1,y2])
        # centered
        if x2-x1>50:
            setproperties(self.ids.ent,x=(x1+x2)/2,y=(y1+y2)/2-10)
            self._ishidden=0
        else:
            # too small- put the widget out of sight (hopefully renders faster)
            if not getattr(self,'_ishidden',0):
                setproperties(self.ids.ent,x=-100,y=0)
                self._ishidden=1
        
        
class Editbrickview(Brickview):
    """a specialized brickview for edit bricks, which includes
    GUI elements for the cuts within the brick"""
    def __init__(self,master,brick):
        self.cutpoints=[] # list of Cutpoint objects that go with our brick

        Brickview.__init__(self,master,brick)

    def withdraw(self):
        Brickview.withdraw(self)
        for cp in self.cutpoints:
            cp.withdraw()

    def _construct(self):
        Brickview._construct(self)
        ids=self.ids
        mas=self.master

        # [cuttime,CanvasLine,Text] lists (should be ClassyDicts). times
        # can be looked up in brick.cuts() mapping. as a cutpoint is edited,
        # the first element (cuttime) is edited.
        ids.cutpoints=[] 

        self.constructcuts()
       
        self.editkeyboardcommands()

        
    def constructcuts(self):
        """rebuilds the cut point UI elements by reading the cuts list from the Brick.
        any old UI elements are properly removed first."""
        ids=self.ids
        mas=self.master

        # remove old items
        while self.cutpoints:
            (self.cutpoints.pop()).withdraw()
            
        # rebuild
        for i,cutpos in enumerate(self.brick.cuts()):
            # the cuts() keys come out sorted
            cp = Cutpoint(self.master,self.brick,cutpos,sensitive=i>0)
            self.cutpoints.append(cp)


    def update(self,fast=0):
        Brickview.update(self,fast)

        for cp in self.cutpoints:
            cp.update()
        

    def editkeyboardcommands(self):
        "adds some more key bindings on the brick"
                
        mapping={ keys.minus : self.newcut,
                  keys.Delete: self.deletecut,
                  keys.exclam: self.dumpbrick,
                  keys.equal: self.newcut, # see special case in handle()
                  }
        def handle(wid,ev):
            if ev.type==gdk.KEY_PRESS:
                if ev.keyval in mapping:
                    x,y,state=ev.window.get_pointer()
                    mousetime=self.master.timeforxcoord(x)
                    if ev.keyval==keys.equal:
                        # equal cuts at timecursor, not mouse
                        mousetime=self.master.cursor.time
                    mapping[ev.keyval](ev,mousetime)
                
            return 0
        
        self.ids.body.connect('event',handle)
    def newcut(self,ev,mousetime):
        """makes a new cut at the time of ev.x"""
        # find what track is already at that point, since that's the track we're going to cut to
        bricktime=mousetime-self.brick.origin()
        cuttrack=self.brick.trackfortime(bricktime)
        # edit model - make new cut
        log.info("cutting edit brick at its time %s" % bricktime)
        self.brick.cuts()[bricktime]=cuttrack
        # reread entire model to notice the new track
        self.constructcuts()
        # and draw
        self.update()
    def dumpbrick(self,ev,mousetime):
        print self.brick.getAttribute("uuid"),self.brick.cuts()
    def deletecut(self,ev,mousetime):
        """removes the cut at the time of the mouse except if it's the only cut"""

        # find the time of the cut we're over (we could be anywhere in the cut)
        if len(self.brick.cuts().keys())>1:
            cutpoint=self.brick.cuts().findcutstart(mousetime-self.brick.origin())
            if cutpoint!=self.brick.cuts().keys()[0]:
                del self.brick.cuts()[cutpoint]
                # redo the cut points on this view to reflect the change
                self.constructcuts()
                self.update()
            else:
                log.warn("please don't delete the first cut point")
