"""
higher-level drawing functions for 2d drawing on an opengl surface.
"""

from OpenGL import GL
from OpenGL.GL import *
if 1:
    import pygame
    pygame.font.init()
else:
    pygame=None

__all__=['setcolor']

def init():
    """call this before using the module."""
    if pygame:
        pygame.init()



def setcolor(color=None):    
    """sets current color to the given r,g,b or r,g,b,a tuple. no
    effect if color is None"""
    if color:
        if len(color)==3:
            glColor3f(*color)
        else:
            glColor4f(*color)

def line(p1,p2,color=None):
    """connects the given points (x,y tuples) in the given color, or
    the current color (r,g,b or r,g,b,a tuple) if none is supplied"""
    setcolor(color)
    glBegin(GL_LINES)
    glVertex2d(*p1)
    glVertex2d(*p2)
    glEnd()

def rect(p1,p2,fillcolor=None,outlinecolor=None):
    """draws a rectangle from the given top-left and
    bottom-right corner points.

    if the fillcolor is supplied, the rect will be filled with it. if
    outlinecolor color is supplied, lines will be drawn around the
    rectangle. the outline happens after the fill.

    colors can be r,g,b or r,g,b,a tuples. the current color is not disturbed.
    """
    glPushAttrib(GL_COLOR_BUFFER_BIT)
    for color,drawmode in (fillcolor,GL_QUADS),(outlinecolor,GL_LINE_LOOP):
        if color:
            setcolor(color)
            glBegin(drawmode)
            glVertex2f(p1[0],p1[1])
            glVertex2f(p1[0],p2[1])
            glVertex2f(p2[0],p2[1])
            glVertex2f(p2[0],p1[1])
            glEnd()
    glPopAttrib()


cachefont={} # (font,size):pygamefont
cachetxtdata={} # (font,size,text,color):image
def text(pos,text,font=None,size=20,color=(1,1,1),anchor='nw'):

    """renders the given text onto the display. text is aligned to x,y
    tuple 'pos' according to the tk-style anchor. text is drawn in the
    specified color, given in opengl floats instead of pygame 8-bit
    channels.

    various font name systems are accepted. that's not settled
    yet. omitting font uses the pygame default font. the font size is
    in pixels, for now. """

    img=text_image(pos,text,font,size,color)
    
    if anchor!='nw':
        raise NotImplementedError("no anchors supported yet besides nw")

    img.draw(pos,_flipy=1)

def text_image(text,font=None,size=20,color=(1,1,1)):
    """like text(), but returns a image for you to place (and read the size)"""

    
    global cachefont,cachetxtdata

    if (font,size) not in cachefont:
        cachefont[(font,size)]=pygame.font.Font(font,size)
    pygamefont=cachefont[(font,size)]

    key=(font,size,text,color)
    if not key in cachetxtdata:
        rendered=pygamefont.render(text,1,[int(c*255) for c in color])
        cachetxtdata[key]=image(rgbastring=pygame.image.tostring(rendered,
                                                                 "RGBA",1),
                                size=rendered.get_size())
        
    img=cachetxtdata[key]
    return img
    
class image:
    
    """an image for drawing onto opengl.

    the image can come from:

      a file (anything pygame/SDL can read);

      an already-read file (anything pygame/SDL can read);

      or a string of RGB or RGBA data.

    supply enough args to the init to describe your data. then use the
    draw method to draw the image."""
    
    def __init__(self,
                 filename=None, # ok by itself
                 filestring=None,filetype="", # or supply these two
                 rgbstring=None,rgbastring=None,size=None # or supply a string and a size
                 ):
        if filename:
            self.fromfilename(filename)
        elif filestring:
            import StringIO
            self.fromfilename(StringIO.StringIO(filestring),filetype)
        elif (rgbstring or rgbastring) and size:
            if rgbstring:
                self.fromstring(rgbstring,size,format="RGB")
            else:
                self.fromstring(rgbastring,size,format="RGBA")
        else:
            raise RuntimeError("see docs for the possible arguments to image()")

        # here's a test of putting the DrawPixels in a calllist. it seems to have
        # not much effect, at least for a little image
        
        if 1:

            # had to switch calllist off when i had multiple GL
            # contexts- only the last-created one would work. and,
            # even when lists are on and the display is broken, it
            # draws slower than when all the text is working

            self.calllist=None
        else:
            self.calllist=glGenLists(1)
            glNewList(self.calllist,GL_COMPILE)
            if 1:
                glformat=getattr(GL,"GL_"+self.format)
                glDrawPixels(self.size[0],self.size[1],
                             glformat,
                             GL_UNSIGNED_BYTE,
                             self.data)
            glEndList()
        
    def fromfilename(self,filename,filetype=""):
        img=pygame.image.load(filename,filetype)
        self.size=img.get_size()
        self.data=pygame.image.tostring(img,"RGBA",1)
        self.format="RGBA"
    def fromstring(self,data,size,format="RGBA"):
        self.size=size
        self.data=data
        self.format=format
    def draw(self,pos,scale=1,_flipy=-1):
        """draws this image at the given pos (starting at
        top-left). image is scaled uniformly by the given factor"""
        # we should probably push more context here

        glRasterPos2d(pos[0],pos[1]+_flipy*self.size[1]*scale)
        glPixelZoom(scale,_flipy*scale)
        if not self.calllist:
            glformat=getattr(GL,"GL_"+self.format)
            glDrawPixels(self.size[0],self.size[1],
                         glformat,
                         GL_UNSIGNED_BYTE,
                         self.data)
        else:
            glCallList(self.calllist)
